namespace ClipboardTool
{
    #region Imports

    using System;

    using Debug = System.Diagnostics.Debug;
    using IEnumerable = System.Collections.IEnumerable;
    using IEnumerator = System.Collections.IEnumerator;

    #endregion

    [ Serializable ]
    internal sealed class CommandLineReader : ICloneable
    {
        private readonly IEnumerator _e;
        private string _text;
        private CommandLineToken _token;

        public CommandLineReader() :
            this(Environment.GetCommandLineArgs()) {}

        public CommandLineReader(IEnumerable args)
        {
            if (args == null)
                args = new string[0];

            _e = args.GetEnumerator();
        }

        private bool ReadNext()
        {
            _text = null;
            return _e.MoveNext();
        }

        public CommandLineToken Read()
        {
            if (ReadNext())
            {
                string name = TryParseName();
                
                if (name.Length > 0)
                    _token = new CommandLineName(name);
                else
                    _token = new CommandLineValue(Text);
            }
            else
            {
                _token = null;
            }

            return _token;
        }

        public CommandLineValue ReadValue()
        {
            if (!ReadNext())
                throw new MissingValueException();

            CommandLineValue value = new CommandLineValue(Text);
            _token = value;
            
            return value;
        }

        public CommandLineToken Token
        {
            get { return _token; }
        }

        public string TryParseName()
        {
            if (Text.Length < 2)
                return string.Empty;

            char ch = this.Text[0];

            if (ch == '-' || ch == '/')
            {
                return Text.Substring(1);
            }
            else
            {
                return string.Empty;
            }
        }

        private string Text
        {
            get 
            { 
                if (_text == null)
                    _text = StringEtc.MaskNull((string) _e.Current);
                
                return _text; 
            }
        }

        public object Clone()
        {
            return MemberwiseClone();
        }
    }

    [ Serializable ]
    public abstract class CommandLineToken
    {
        private string _text;

        protected CommandLineToken(string text)
        {
            _text = text;
        }

        public virtual string Text
        {
            get { return StringEtc.MaskNull(_text); }
        }

        public override string ToString()
        {
            return Text;
        }
    }

    [ Serializable ]
    public sealed class CommandLineName : CommandLineToken
    {
        public CommandLineName(string text) :
            base(text) {}
    }

    [ Serializable ]
    public sealed class CommandLineValue : CommandLineToken, IConvertible
    {
        public CommandLineValue(string text) :
            base(text) {}

        //
        // IConvertible implementation
        //

        public TypeCode GetTypeCode() { return TypeCode.Object; }
        [ CLSCompliant(false) ] public ulong ToUInt64(IFormatProvider provider) { return Convert.ToUInt64(Text, provider); }
        [ CLSCompliant(false) ] public sbyte ToSByte(IFormatProvider provider) { return Convert.ToSByte(Text, provider); }
        public double ToDouble(IFormatProvider provider) { return Convert.ToDouble(Text, provider); }
        public DateTime ToDateTime(IFormatProvider provider) { return Convert.ToDateTime(Text, provider); }
        public float ToSingle(IFormatProvider provider) { return Convert.ToSingle(Text, provider); }
        public bool ToBoolean(IFormatProvider provider) { return Convert.ToBoolean(Text, provider); }
        public int ToInt32(IFormatProvider provider) { return Convert.ToInt32(Text, provider); }
        [ CLSCompliant(false) ] public ushort ToUInt16(IFormatProvider provider) { return Convert.ToUInt16(Text, provider); }
        public short ToInt16(IFormatProvider provider) { return Convert.ToInt16(Text, provider); }
        public string ToString(IFormatProvider provider) { return Convert.ToString(Text, provider); }
        public byte ToByte(IFormatProvider provider) { return Convert.ToByte(Text, provider); }
        public char ToChar(IFormatProvider provider) { return Convert.ToChar(Text, provider); }
        public long ToInt64(IFormatProvider provider) { return Convert.ToInt64(Text, provider); }
        public decimal ToDecimal(IFormatProvider provider) { return Convert.ToDecimal(Text, provider); }
        public object ToType(Type conversionType, IFormatProvider provider) { return Convert.ChangeType(Text, conversionType, provider); }
        [ CLSCompliant(false) ] public uint ToUInt32(IFormatProvider provider) { return Convert.ToUInt32(Text, provider); }
    }
}
