namespace ClipboardTool
{
    #region Imports
    
    using System;
    
    using HttpUtility = System.Web.HttpUtility;
    using IDataObject = System.Windows.Forms.IDataObject;
    using DataObject = System.Windows.Forms.DataObject;
    using DataFormats = System.Windows.Forms.DataFormats;
    using StringBuilder = System.Text.StringBuilder;
    using StringWriter = System.IO.StringWriter;
    using AssemblyName = System.Reflection.AssemblyName;
    using Regex = System.Text.RegularExpressions.Regex;
    using RegexOptions = System.Text.RegularExpressions.RegexOptions;
    using Match = System.Text.RegularExpressions.Match;
    using Encoding = System.Text.Encoding;
    using TextReader = System.IO.TextReader;
    using TextWriter = System.IO.TextWriter;
    using Debug = System.Diagnostics.Debug;
    using DllImportAttribute = System.Runtime.InteropServices.DllImportAttribute;

    using HWND = System.IntPtr;
    using HANDLE = System.IntPtr;
    using HGLOBAL = System.IntPtr;
    using LPVOID = System.IntPtr;
    using UINT = System.UInt32;
    using BOOL = System.Boolean;
    using Marshal = System.Runtime.InteropServices.Marshal;
    using Win32Exception = System.ComponentModel.Win32Exception;
    using MemoryStream = System.IO.MemoryStream;

    #endregion

    internal sealed class HtmlHandler : IDataObjectHandler
	{
        public IDataObject Copy(TextReader reader)
        {
            Debug.Assert(reader != null);

            string text = reader.ReadToEnd();

            //
            // Assume that the text is pre-formatted (PRE) and encode it
            // according to HTML rules.
            // 

            string htmlFragment = "<PRE>" + HttpUtility.HtmlEncode(text) + "</PRE>";

            //
            // Get the containing assembly name, which will be used as the
            // title of the HTML document.
            //

            AssemblyName thisAssemblyName = this.GetType().Assembly.GetName();

            //
            // Write out the HTML Clipboard (CF_HTML) format. For details on 
            // this format, see the following documentation in MSDN:
            // http://msdn.microsoft.com/workshop/networking/clipboard/htmlclipboard.asp
            //
            // As the format is written out, we note offsets that delimit
            // the HTML document as well as the embedded fragment. Since all
            // offsets go into the format header and we cannot possibly know 
            // them in  advance, we leave fixed-length markers that are 
            // replaced upon completion.
            //

            StringBuilder html = new StringBuilder(htmlFragment.Length + 500);
            StringWriter htmlWriter = new StringWriter(html);

            const string htmlStartOffsetMarker = "SHOXXXXXXX";
            const string htmlEndOffsetMarker = "EHOXXXXXXX";
            const string fragmentStartOffsetMarker = "SFOXXXXXXX";
            const string fragmentEndOffsetMarker = "EFOXXXXXXX";

            htmlWriter.WriteLine("Version:0.9");
            htmlWriter.Write("StartHTML:");
            htmlWriter.WriteLine(htmlStartOffsetMarker);
            htmlWriter.Write("EndHTML:");
            htmlWriter.WriteLine(htmlEndOffsetMarker);
            htmlWriter.Write("StartFragment:");
            htmlWriter.WriteLine(fragmentStartOffsetMarker);
            htmlWriter.Write("EndFragment:");
            htmlWriter.WriteLine(fragmentEndOffsetMarker);
            int htmlStartOffset = html.Length;
            htmlWriter.WriteLine("<HTML>");
            htmlWriter.WriteLine("<TITLE>{0} V{1}</TITLE>", 
                HttpUtility.HtmlEncode(thisAssemblyName.Name), 
                thisAssemblyName.Version);
            htmlWriter.WriteLine("<BODY>");
            htmlWriter.Write("<!--StartFragment-->");
            htmlWriter.Flush();
            int fragmentStartOffset = html.Length;
            htmlWriter.WriteLine();
            htmlWriter.WriteLine(htmlFragment);
            int fragmentEndOffset = html.Length;
            htmlWriter.WriteLine("<!--EndFragment-->");
            htmlWriter.WriteLine("</BODY>");
            htmlWriter.WriteLine("</HTML>");
            int htmlEndOffset = html.Length;

            //
            // Now go back and patch the header with the known offset.
            //

            const string offsetFormat = "0000000000";

            Debug.Assert(htmlStartOffsetMarker.Length == offsetFormat.Length);
            Debug.Assert(htmlEndOffsetMarker.Length == offsetFormat.Length);
            Debug.Assert(fragmentStartOffsetMarker.Length == offsetFormat.Length);
            Debug.Assert(fragmentEndOffsetMarker.Length == offsetFormat.Length);

            html.Replace(htmlStartOffsetMarker, htmlStartOffset.ToString(offsetFormat), 0, htmlStartOffset);
            html.Replace(htmlEndOffsetMarker, htmlEndOffset.ToString(offsetFormat), 0, htmlStartOffset);
            html.Replace(fragmentStartOffsetMarker, fragmentStartOffset.ToString(offsetFormat), 0, htmlStartOffset);
            html.Replace(fragmentEndOffsetMarker, fragmentEndOffset.ToString(offsetFormat), 0, htmlStartOffset);

            //
            // Finally, create the data object representing the format
            // and return it.
            //

            DataObject dataObject = new DataObject();
            dataObject.SetData(DataFormats.Html, html.ToString());

            return dataObject;
        }

        public bool Paste(IDataObject source, TextWriter writer)
        {
            Debug.Assert(source != null);
            Debug.Assert(writer != null);

            if (!source.GetDataPresent(DataFormats.Html))
                return false;

            //
            // Get the HTML clipboard format data as a string.
            //
            // NOTE: The WinForms implementation in .NET Framework v1.0.3705 
            // has a bug that incorrectly treats the HTML clipboard format
            // as a raw byte array and converts it to a string without 
            // applying the UTF-8 decoding rules. When dealing with the 
            // header information, this is not a problem, but to read the
            // actual context and fragment, we will later have to reconvert
            // the data.
            //

            byte[] data = GetClipboardHtml();
            string clipboardHtml = Encoding.UTF8.GetString(data);

            //
            // Get the start and end offsets of the context.
            //

            int contextStartOffset = GetHeaderOffset(clipboardHtml, "StartHTML");
            int contextEndOffset = GetHeaderOffset(clipboardHtml, "EndHTML");
            int contextLength = contextEndOffset - contextStartOffset;

            try
            {
                string html = Encoding.UTF8.GetString(data, contextStartOffset, contextLength);
                writer.Write(html);
            }
            catch (ArgumentOutOfRangeException)
            {
                throw new ProgramException("Data is not valid clipboard HTML format.");
            }

            return true;
        }

        private static int GetHeaderOffset(string header, string offsetName)
        {
            Match match = Regex.Match(header, offsetName + @":(\d+)", RegexOptions.IgnoreCase);

            if (!match.Success)
                throw new ProgramException(string.Format("Data is not valid clipboard HTML format. {0} was not found.", offsetName));

            try
            {
                return int.Parse(match.Groups[1].Value);
            }
            catch (FormatException)
            {
                throw new ProgramException(string.Format("Data is not valid clipboard HTML format."));
            }
        }

        private static byte[] GetClipboardHtml()
        {
            //
            // Open the clipboard.
            //

            if (!Native.OpenClipboard(HWND.Zero))
            {
                Win32Exception e = new Win32Exception();
                throw new ProgramException("Unable to open the clipboard.", e);
            }
    
            try
            {
                //
                // Get the HTML Format data from the clipboard.
                //
                // NOTE: For some odd reason, GetClipboardData fails under
                // the debugger with the system error, "Thread does not have 
                // a clipboard open." This happens even though the call
                // to OpenClipboard succeeds.
                //

                int formatAtom = DataFormats.GetFormat(DataFormats.Html).Id;
                
                HGLOBAL dataHandle = Native.GetClipboardData((UINT) formatAtom);

                if (dataHandle == HGLOBAL.Zero)
                {
                    Win32Exception e = new Win32Exception();
                    throw new ProgramException("Unable to acquire clipboard data.", e);
                }

                //
                // Lock the memory handle to get a pointer to the
                // raw unmanaged memory.
                //

                LPVOID data = Native.GlobalLock(dataHandle);

                if (data == LPVOID.Zero)
                {
                    Win32Exception e = new Win32Exception();
                    throw new ProgramException("Unable to lock clipboard data.", e);
                }
                
                try
                {
                    //
                    // Copy the unmanaged memory to a managed buffer, 
                    // byte-by-byte. This is probably very slow and could be
                    // speeded up with use of unsafe code. Also, we use
                    // a memory-based stream because we don't know how
                    // big the data could be. A possible improvement here 
                    // could be to use GlobalSize to get an estimate of the
                    // buffer size and use that to set the capacity of
                    // the memory stream.
                    //

                    MemoryStream dataStream = new MemoryStream();

                    int count = 0;
                    byte oneByte;

                    while ((oneByte = Marshal.ReadByte(data, count++)) != 0)
                        dataStream.WriteByte(oneByte);

                    dataStream.Close();
                    return dataStream.ToArray();
                }
                finally
                {
                    Native.GlobalUnlock(dataHandle);
                }
            }
            finally
            {
                Native.CloseClipboard();
            }
        }

        private sealed class Native
        {
            [ DllImport("user32.dll", SetLastError = true) ]
            public static extern BOOL OpenClipboard(HWND hWndNewOwner);

            [ DllImport("user32.dll", SetLastError = true) ]
            public static extern BOOL CloseClipboard();

            [ DllImport("user32.dll", SetLastError = true) ]
            public static extern HANDLE GetClipboardData(UINT uFormat);
        
            [ DllImport("kernel32.dll", SetLastError = true) ]
            public static extern LPVOID GlobalLock(HGLOBAL handle);

            [ DllImport("kernel32.dll", SetLastError = true) ]
            public static extern BOOL GlobalUnlock(HGLOBAL handle);

            private Native() {}
        }
    }
}
